# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: Messager.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eMessager.proto\"\x1a\n\x07Message\x12\x0f\n\x07message\x18\x01 \x01(\t\"/\n\x06Server\x12\x16\n\x0eidentification\x18\x01 \x01(\t\x12\r\n\x05token\x18\x03 \x01(\x0c\"\x06\n\x04Void2y\n\tMessenger\x12%\n\x11RequestConnection\x12\x07.Server\x1a\x05.Void\"\x00\x12 \n\x0c\x41uthenticate\x12\x07.Server\x1a\x05.Void\"\x00\x12#\n\x07\x43onnect\x12\x08.Message\x1a\x08.Message\"\x00(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'Messager_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MESSAGE']._serialized_start=18
  _globals['_MESSAGE']._serialized_end=44
  _globals['_SERVER']._serialized_start=46
  _globals['_SERVER']._serialized_end=93
  _globals['_VOID']._serialized_start=95
  _globals['_VOID']._serialized_end=101
  _globals['_MESSENGER']._serialized_start=103
  _globals['_MESSENGER']._serialized_end=224
# @@protoc_insertion_point(module_scope)
